"""
Module for data engineeringt
"""

import geopandas as gpd
import pandas as pd
import tarfile
import re
import datetime
import json

from .geo_utils import get_image_bbox
from shapely.geometry import box
from os.path import exists
from typing import Union, Optional


def get_images_by_location(gdf: gpd.GeoDataFrame) -> pd.DataFrame:
    """
    Generate a GeoDataFrame with the available images for each location in the dataset. 

    :param gdf: GeoDataFrame generated from the ItemCollection of a sen12floods collection
    :return gdf_dates_per_aoi:  GeoDataFrame with the available images for each location in
                                the dataset.
            The returned GeoDataFrame has three columns:
                - location_id: the unique ID of each location.
                - images_count: the count of available images of each location.
                - images_dates: list with the dates of the available images of each location.
    """
    uniques_location_id = gdf['scene_id'].unique()   # List of unique location ids
    uniques_location_id.sort()

    images_count_list, images_dates_list = [], []

    # Iterate the unique location ids, count the number of images per location and generate
    # a list with the dates of every image in a location
    for location_id in uniques_location_id:
        dates = gdf[gdf['scene_id'] == location_id]['datetime']
        images_count_list.append(dates.count())
        images_dates_list.append(dates.tolist())

    images_dates_list.sort()   # Sort the list of dates
    data = {'scene_id': uniques_location_id, 'dates_count': images_count_list, 'dates_list': images_dates_list}
    df_dates_per_aoi = pd.DataFrame.from_dict(data)

    return df_dates_per_aoi


def generate_location_payload(gdf: Union[gpd.GeoDataFrame, pd.DataFrame], path: str) -> dict:
    """
    Generate a dictionary with the location payload of the locations in the GeoDataFrame, 
    such as the bounding box and the time interval to search for available data.
    """
    payload_cache = f"{path}/location_payload.json"
    if exists(payload_cache):
        # Read as dict
        with open(payload_cache, 'r') as f:
            payload = json.load(f)
        return payload
    
    bbox_date_by_location = dict()
    for i, row in gdf.iterrows():
        # Get list from dates_list column
        dates_list = list(row['dates_list'])
        for date in dates_list:
            location_id = row['location_id']
            date_formatted = datetime.datetime.strptime(date, '%Y-%m-%dT%H:%M:%S.%fZ').strftime('%Y-%m-%d')
            location_id_formatted = f'{location_id}_{date_formatted}'
            bbox_date_by_location[location_id_formatted] = {
                'bounding_box': row['geometry'].bounds,
                # Convert str to datetime
                'time_interval': (date, date)
            }

    # Save to json
    with open(payload_cache, 'w') as f:
        json.dump(bbox_date_by_location, f)

    return bbox_date_by_location


def get_tarfile_image_info(tar: str, path: Optional[str] = None, pattern: Optional[str] = r"\d{8}T\d{6}", level: Optional[int] = 2):
    """
    """
    if path:
        gdf_cache = f"{path}/tarfile_images_info.csv"
        if exists(gdf_cache):
            images_gdf = gpd.read_file(gdf_cache,
                                       GEOM_POSSIBLE_NAMES="geometry", 
                                       KEEP_GEOM_COLUMNS="NO")
            images_gdf.set_crs(epsg=4326, inplace=True)
            
            return images_gdf
    
    images_df = pd.DataFrame()
    with tarfile.open(tar, 'r:gz') as tar:
        rasters = [i for i in tar.getnames() if i.endswith(".tif") or i.endswith(".tiff")]
        for raster in rasters:
            r = tar.extractfile(raster)
            bbox = get_image_bbox(r)
            date = extract_image_date_in_folder(raster, pattern)
            date_formatted = datetime.datetime.strptime(date, '%Y-%m-%dT%H:%M:%S.%fZ').strftime('%Y-%m-%d')
            id = extract_image_id_in_folder(raster, level)
            # Use pd.concat to append to dataframe
            images_df = pd.concat([images_df, pd.DataFrame({"location_id": [id], 
                                                            "datetime": [date], 
                                                            "bbox": [bbox]})])

    # Clean duplicates
    images_df = images_df.drop_duplicates(subset=["location_id", "datetime"])
    # Convert to geodataframe
    images_gdf = gpd.GeoDataFrame(images_df, 
                                  crs='EPSG:4326',
                                  geometry=images_df["bbox"].apply(lambda x: box(x[0], x[1], x[2], x[3])))
    # Drop bbox column
    images_gdf = images_gdf.drop(columns=["bbox"])
    # Set crs
    images_gdf = images_gdf.set_crs(epsg=4326)
    # Sort by location_id
    images_gdf = images_gdf.sort_values(by=["location_id"])
    if path:
        # Save to csv
        images_gdf.to_csv(gdf_cache, index=False)

    return images_gdf


def extract_image_date_in_folder(raster_path: str, pattern: str):
    case = re.findall(pattern, raster_path)

    if case:
        date = case[0]
        # Convert date to format YYYY-MM-DDT00:00:00.000Z as datetime object
        formatted_date = datetime.datetime.strptime(date, '%Y%m%dT%H%M%S').strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        return formatted_date
    
    return None


def extract_image_id_in_folder(raster_path: str, level: int):
    return raster_path.split("/")[level]


def format_product_location_payload(location_payload: dict,
                                    images_response: dict,
                                    all_info: bool = False
                                    ) -> dict:
    """
    """
    for id, info in location_payload.items():
        # Add new key to the dictionary
        if all_info:
            location_payload[id]['image'] = images_response[id] if id in images_response else None
        else:
            location_payload[id]['image'] = images_response[id]['properties']['id'] if images_response[id] else None

    return location_payload



"""
Utils
"""

from sentinelhub import DataCollection, MosaickingOrder

from .evalscripts import EvalScripts


class SHParameters:

    def __init__(self):
        pass


class SHS2L2AParameters(SHParameters):
    DATA_COLLECTION = DataCollection.SENTINEL2_L2A
    RESOLUTION = 10
    MOSAICKING_ORDER = MosaickingOrder.LEAST_CC
    EVALSCRIPT = EvalScripts.SENTINEL_2_L2A
    FIELDS = {"include": ["id", 
                        "properties.datetime",
                        "properties.eo:cloud_cover"], 
            "exclude": []}
    FILTER = None


class SHS2L1CParameters(SHParameters):
    DATA_COLLECTION = DataCollection.SENTINEL2_L1C
    RESOLUTION = 10
    MOSAICKING_ORDER = MosaickingOrder.LEAST_CC
    EVALSCRIPT = EvalScripts.SENTINEL_2_L1C
    FIELDS = {"include": ["id", 
                        "properties.datetime",
                        "properties.eo:cloud_cover"], 
            "exclude": []}
    

class SHS1Parameters(SHParameters):
    DATA_COLLECTION = DataCollection.SENTINEL1
    RESOLUTION = 3
    EVALSCRIPT = EvalScripts.SENTINEL_1
    FIELDS = {"include": ["id", 
                        "properties.datetime",
                        "sar:instrument_mode", 
                        "s1:polarization",
                        "sat:orbit_state",
                        "s1:resolution",
                        "s1:timeliness"], 
            "exclude": []},
    FILTER = None


class SHDEMParameters(SHParameters):
    DATA_COLLECTION = DataCollection.DEM_COPERNICUS_30
    RESOLUTION = 3
    MOSAICKING_ORDER = None
    EVALSCRIPT = EvalScripts.DEM
    FILTER = None
    FIELDS = None


SUPPORTED_SENSORS = ('sentinel-1-grd', 'sentinel-2-l1c', 'sentinel-2-l2a', 'dem')

SH_PARAMETERS_DICT = {
    'sentinel-1-grd': SHS1Parameters,
    'sentinel-2-l1c': SHS2L1CParameters,
    'sentinel-2-l2a': SHS2L2AParameters,
    'dem': SHDEMParameters
}



from .sentinelhub import SHClient, SH_PARAMETERS_DICT, evaluate_sentinel_parameters, imagery_from_tmp_to_dir
from .search import search_sentinel_imagery

from shutil import rmtree
from datetime import datetime
from typing import Union, List


def download_sentinel_imagery(output: str,
                              time_interval: Union[str, datetime, List[Union[str, datetime]]],
                              bounding_box: List[Union[int, float]],
                              sensor: str,
                              ) -> None:
    evaluate_sentinel_parameters(sensor, time_interval, bounding_box, output)

    client = SHClient()
    parameters = SH_PARAMETERS_DICT[sensor]()

    request = client.request_data(time_interval, bounding_box, parameters)
    client.download_data(request)
    imagery_from_tmp_to_dir(output)


def search_and_download_sentinel_imagery(output: str,
                              time_interval: Union[str, datetime, List[Union[str, datetime]]],
                              bounding_box: List[Union[int, float]],
                              sensor: str,
                              ) -> None:
    evaluate_sentinel_parameters(sensor, time_interval, bounding_box, output)

    client = SHClient()
    parameters = SH_PARAMETERS_DICT[sensor]()

    results = search_sentinel_imagery(time_interval, bounding_box, sensor)
    timestamps = [date.strftime("%Y-%m-%d") for date in results.get_timestamps()]

    requests_list = list()
    for date in timestamps:
        requests_list.append(client.request_data(date, bounding_box, parameters))
    client.download_data(requests_list)
    imagery_from_tmp_to_dir(output)



from .sentinelhub import SHClient, SH_PARAMETERS_DICT, evaluate_sentinel_parameters
from typing import Union, List
from datetime import datetime


def search_sentinel_imagery(time_interval: Union[str, datetime, List[Union[str, datetime]]],
                            bounding_box: List[Union[int, float]],
                            sensor: str,
                            ) -> None:
    evaluate_sentinel_parameters(sensor, time_interval, bounding_box, output_needed=False)

    client = SHClient()
    parameters = SH_PARAMETERS_DICT[sensor]()

    return client.search_data(bounding_box, time_interval, parameters)



"""
Module for managing the Sentinel Hub configuration and data access
"""

import json
from os import getenv
from os.path import exists
from sentinelhub import (SHConfig, 
                         SentinelHubCatalog, 
                         BBox, 
                         bbox_to_dimensions, 
                         CRS, 
                         SentinelHubRequest, 
                         SentinelHubDownloadClient,
                         MimeType)

from ...repos.AuthRepo import AuthRepo
from .parameters import SHParameters
from ...tools.time_utils import prepare_time_interval
from ...tools.geo_utils import compute_image_size


class SHClient():
    """
    Client class to manage the Sentinel Hub Python interface.
    """

    def __init__(self
                 ) -> None:
        """
        """
        self.config = SHConfig()
        if getenv('SH_CLIENT_ID') and getenv('SH_CLIENT_SECRET'):
            # If the user has provided the credentials in a .env file, we should save them
            self.config.sh_client_id = getenv('SH_CLIENT_ID')
            self.config.sh_client_secret = getenv('SH_CLIENT_SECRET')
        else:
            # If the user has not provided the credentials, we should check if
            # the user is logged in
            auth_repo = AuthRepo()
            creds_file = auth_repo.creds_path
            if not exists(creds_file):
                raise ValueError('Your are not logged in and have not provided Sentinel Hub credentials. Please, crete a .env file with your SH_CLIENT_ID and SH_CLIENT_SECRET, or login')
            else:
                creds = json.load(open(creds_file, 'r'))
                if not 'SH_CLIENT_ID' in creds.keys() or not 'SH_CLIENT_SECRET' in creds.keys():
                    raise ValueError('If you already had a Sentinel HUB account before accepting the Terms and Conditions, your SH credentials will NOT appear here. You can retrieve them from you Sentinel HUB dashboard)')
                else:
                    self.config.sh_client_id = creds['SH_CLIENT_ID']
                    self.config.sh_client_secret = creds['SH_CLIENT_SECRET']
        self.catalog = SentinelHubCatalog(config=self.config)
        self.tmp_dir = '/tmp/sentinelhub' 

    def search_data(self,
                    bounding_box: list,
                    time_interval: list,
                    parameters: SHParameters
                    ) -> list:
        """
        """
        search_iterator = self.catalog.search(
            parameters.DATA_COLLECTION,
            bbox=BBox(bounding_box, crs=CRS.WGS84),
            time=time_interval,
            filter=parameters.FILTER,
            fields=parameters.FIELDS
        )

        return search_iterator
    
    def request_data(self,
                     time_interval,
                     bounding_box: list,
                     parameters: SHParameters
                     ) -> list:
        """
        """
        time_interval = prepare_time_interval(time_interval)
        bounding_box, bounding_box_size = compute_image_size(bounding_box, parameters)

        return SentinelHubRequest(
                    data_folder=self.tmp_dir,
                    evalscript=parameters.EVALSCRIPT,
                    input_data=[
                        SentinelHubRequest.input_data(
                            data_collection=parameters.DATA_COLLECTION,
                            time_interval=time_interval,
                            mosaicking_order=parameters.MOSAICKING_ORDER,
                        )
                    ],
                    responses=[
                                SentinelHubRequest.output_response("default", MimeType.TIFF)
                    ],
                    bbox=bounding_box,
                    size=bbox_to_dimensions(bounding_box, parameters.RESOLUTION),
                    config=self.config,
                )

    def download_data(self,
                      requests: list
                      ) -> list:
        """
        """
        download_client = SentinelHubDownloadClient(config=self.config)
        if not isinstance(requests, list):
            requests = [requests]
        download_requests = [request.download_list[0] for request in requests]
        data = download_client.download(download_requests)

        return data



import geopandas as gpd
import rasterio
import rasterio.warp
import tarfile

from typing import Union
from shapely import geometry
from shapely.geometry import box, Polygon, shape
from pyproj import Transformer
from sentinelhub import BBox, CRS, bbox_to_dimensions
from pandas import isna


def is_bounding_box(bbox: list) -> bool:
    if not isinstance(bbox, (list, tuple)) or len(bbox) != 4:
        return False
    
    for value in bbox:
        if not isinstance(value, (int, float)):
            return False

    minx, miny, maxx, maxy = bbox
    if minx >= maxx or miny >= maxy:
        return False

    return True


def compute_image_size(bounding_box, parameters):
    bbox = BBox(bbox=bounding_box, crs=CRS.WGS84)
    bbox_size = bbox_to_dimensions(bbox, resolution=parameters.RESOLUTION)

    return bbox, bbox_size  


def get_image_bbox(raster: Union[tarfile.ExFileObject, str]):
    with rasterio.open(raster) as src:
        bounds = src.bounds
        dst_crs = "EPSG:4326"
        left, bottom, right, top = rasterio.warp.transform_bounds(
            src.crs, dst_crs, *bounds
        )
        bbox = [left, bottom, right, top]
    return bbox


def get_image_resolution(raster: Union[tarfile.ExFileObject, str]):
    with rasterio.open(raster) as src:
        resolution = src.res
    return resolution


def bbox_to_coordinates(bounding_box: list) -> list:
    """
    """
    polygon_coordinates = [
        (bounding_box[0], bounding_box[1]),  # bottom left
        (bounding_box[0], bounding_box[3]),  # top left
        (bounding_box[2], bounding_box[3]),  # top right
        (bounding_box[2], bounding_box[1]),  # bottom right
        (bounding_box[0], bounding_box[1])   # back to bottom left
    ]

    return polygon_coordinates


def bbox_to_polygon(bounding_box: list) -> Polygon:
    """
    """
    polygon = box(bounding_box[0], bounding_box[1], bounding_box[2], bounding_box[3])

    return polygon


from_4326_transformer = Transformer.from_crs('EPSG:4326', 'EPSG:3857')
from_3857_transformer = Transformer.from_crs('EPSG:3857', 'EPSG:4326')


def bbox_from_centroid(x: Union[int, float], 
                       y: Union[int, float],
                       pixel_size: Union[int, float],
                       width: Union[int, float],
                       height: Union[int, float]
                       ) -> list:
    """
    Generate a bounding box from a centroid, pixel size and image dimensions.

    Params
    ------
    x: int or float
        x coordinate of the centroid
    y: int or float
        y coordinate of the centroid
    pixel_size: int or float
        pixel size in meters
    width: int or float
        width of the image in pixels
    height: int or float
        height of the image in pixels

    Returns
    -------
    bounding_box: list
        list with the bounding box coordinates
    """
    width_m = width * pixel_size
    heigth_m = height * pixel_size

    # Transform the centroid coordinates to meters
    centroid_m = from_4326_transformer.transform(x, y)

    # Calculate the bounding box coordinates
    min_x = centroid_m[0] - width_m / 2
    min_y = centroid_m[1] - heigth_m / 2
    max_x = centroid_m[0] + width_m / 2
    max_y = centroid_m[1] + heigth_m / 2

    # Convert the bounding box coordinates back to degrees
    min_x, min_y = from_3857_transformer.transform(min_x, min_y)
    max_x, max_y = from_3857_transformer.transform(max_x, max_y)

    return [min_y, min_x, max_y, max_x]


def generate_bounding_box(geom: geometry.point.Point, 
                          differences: list
                          ) -> list:
    """
    Generate the bounding box of a given point using the difference
    between the maximum and mininum coordinates of the bounding box

    :param geom: shapely geometry object of the point which we want to
                generate the bounding box.
    :param differences: list with the difference between the maximum
                and minimum longitude and latitude coordinates.
    :return: list with the resulting bounding box from the computing.
    """
    long_diff, lat_diff = differences[0], differences[1]
    lon, lat = geom.x, geom.y
    
    bbox = (lon - (long_diff/2), 
            lat - (lat_diff/2), 
            lon + (long_diff/2), 
            lat + (lat_diff/2))
    
    # Round the coordinates to 6 decimals
    bounding_box = [round(i, 6) for i in bbox]

    return bounding_box


def calculate_average_coordinates_distance(bounding_box_by_location: dict) -> list:
    """
    Calculate the mean distance between maximum and minixum longitude and latitude of the bounding boxes
    from the existing locations. This is intended to use these mean distance to generate the bounding 
    boxes of the new locations given a centroid.

    :param bounding_box_by_location: dictionary with format location_id : bounding_box for the existing
            locations in the sen12floods dataset.
    :return mean_long_diff, mean_lat_diff: mean longitude and latitude difference in the bounding boxes
    """
    from statistics import mean

    long_diff_list, lat_diff_list = list(), list()

    for bbox in bounding_box_by_location.values():
        long_diff = bbox[2] - bbox[0]
        long_diff_list.append(long_diff)
        lat_diff = bbox[3] - bbox[1]
        lat_diff_list.append(lat_diff)

    mean_long_diff = mean(long_diff_list)
    mean_lat_diff = mean(lat_diff_list)

    return mean_long_diff, mean_lat_diff


def generate_new_locations_bounding_boxes(gdf: gpd.GeoDataFrame,
                                          mean_differences: list,
                                          latest_id: int
                                          ) -> dict:
    """
    Generate the bounding box of every new location, using the mean difference between the maximum and
    minimum calculated longitude and latitude. This function also returns the time interval which we
    want to request from Sentinel Hub Services.

    :param gdf: GeoDataFrame wiht the new locations that are going to be added to the dataset
    :param mean_differences: list with the longitude and latitude mean differences, which are going to be used
            to generate the bounding boxes.
    :return: bbox_by_new_location: dict with format {<location_id>: {'bounding_box': list(), 'time_interval': list()}, ... }
            that contains the bounding box and time interval of the imagery for each location
    """
    bbox_by_new_location = dict()

    for i, row in gdf.iterrows():
        new_location_id = str(latest_id + 1)
        time_interval = row['Began'].strftime("%Y-%m-%d"), row['Ended'].strftime("%Y-%m-%d")
        bbox = generate_bounding_box(row['geometry'], mean_differences)
        bbox_by_new_location[new_location_id] = {'bounding_box': bbox, 'time_interval': time_interval}
        latest_id += 1

    return bbox_by_new_location


def convert_df_geom_to_shape(row):
    """
    Convert the geometry of a dataframe row to a shapely shape
    """
    if not isna(row["geometry"]):
        geo = shape(row["geometry"])
        wkt = geo.wkt
    else:
        wkt = "POLYGON EMPTY"

    return wkt



from .download import (download_sentinel_imagery,
                       search_and_download_sentinel_imagery)
from .search import search_sentinel_imagery
from .sentinelhub.parameters import SUPPORTED_SENSORS


from pathlib import Path
import yaml

from ..auth import with_auth
from .metadata import Metadata
from ..repos import DatasetsAPIRepo
from ..files import ingest_files


def ingest_dataset(path, verbose=False, logger=print):
    path = Path(path)
    if not path.is_dir():
        raise Exception("Path must be a folder")
    # if "catalog.json" in [f.name for f in path.iterdir()]:
    #     return ingest_stac(path / "catalog.json", logger)
    return ingest_folder(path, verbose, logger)


def retrieve_dataset(metadata, user):
    repo = DatasetsAPIRepo()
    data, error = repo.retrieve_dataset(metadata.name)
    # print(data, error)
    if data and data["uid"] != user["sub"]:
        raise Exception("Dataset already exists.")
    if error and error == "Dataset doesn't exist":
        # create dataset
        data, error = repo.create_dataset(metadata.dict(), user["id_token"])
        # print(data, error)
        if error:
            raise Exception(error)
        data["id"] = data["dataset_id"]
    return data["id"]


@with_auth
def ingest_folder(folder, verbose=False, logger=print, user=None):
    repo = DatasetsAPIRepo()
    # load metadata
    metadata = yaml.safe_load(open(folder.joinpath("metadata.yml"), "r").read()) or {}
    metadata = Metadata(**metadata)
    # retrieve dataset (create if doesn't exist)
    dataset_id = retrieve_dataset(metadata, user)
    # ingest files
    return ingest_files(
        repo, dataset_id, folder, verbose, logger, user, endpoint="datasets"
    )


# @with_auth
# def ingest_stac(stac_catalog, logger=None, user=None):
#     api_repo = APIRepo()
#     ingest = IngestSTAC(api_repo, ingest_file, logger)
#     inputs = ingest.Inputs(stac_catalog=stac_catalog, user=user)
#     outputs = ingest(inputs)
#     return outputs.dataset
